from django.urls import path

from . import views

app_name = "visitors"

urlpatterns = [
    path(
        "self-service/<uuid:visitor_uuid>/",
        views.SelfService.as_view(),
        name="self-service",
    ),
    path(
        "self-service/<uuid:visitor_uuid>/success/",
        views.self_service_success,
        name="self-service-success",
    ),
]



from __future__ import annotations

import logging
from typing import Callable

from django.conf import settings
from django.core.exceptions import MiddlewareNotUsed, ValidationError
from django.http.request import HttpRequest
from django.http.response import HttpResponse, HttpResponseBadRequest

from . import session
from .models import InvalidVisitorPass, Visitor
from .settings import VISITOR_QUERYSTRING_KEY

logger = logging.getLogger(__name__)


class VisitorRequestMiddleware:
    """Extract visitor token from incoming request."""

    def __init__(self, get_response: Callable):
        self.get_response = get_response

    def __call__(self, request: HttpRequest) -> HttpResponse:
        request.visitor = None
        request.user.is_visitor = False
        visitor_uuid = request.GET.get(VISITOR_QUERYSTRING_KEY)
        if not visitor_uuid:
            return self.get_response(request)
        try:
            visitor = Visitor.objects.get(uuid=visitor_uuid)
            visitor.validate()
        except Visitor.DoesNotExist:
            logger.debug("Visitor pass does not exist: %s", visitor_uuid)
            return self.get_response(request)
        except InvalidVisitorPass as ex:
            logger.debug("Invalid access request: %s", ex)
            return self.get_response(request)
        except ValidationError as ex:
            logger.debug("Malformed visitor token: %s", ex)
            return HttpResponseBadRequest("Malformed visitor token.")
        else:
            request.visitor = visitor
            request.user.is_visitor = True
        return self.get_response(request)


class VisitorSessionMiddleware:
    """Extract visitor info from session and update request user."""

    def __init__(self, get_response: Callable):
        self.get_response = get_response

    def __call__(self, request: HttpRequest) -> HttpResponse:
        """
        Update request.user if any visitor vars are found in session.

        This middleware should run after the VisitorRequestMiddleware. If
        a request comes in without a visitor token in the querystring, we
        need to check the request.session to see if any visitor data was
        stashed previously.

        The first time this middleware runs after a new session is created
        this will push the `request.visitor` info into the session.
        Subsequent requests will then get the data out of the session.

        """
        # This will only be true directly after VisitorRequestMiddleware
        # has set the values. All subsequent requests in the session will
        # start with is_visitor=False and pick up the visitor info from
        # the session.
        if request.visitor:
            session.stash_visitor_uuid(request)
            return self.get_response(request)

        # We don't have a visitor object, but there may be one in the session
        if not (visitor_uuid := session.get_visitor_uuid(request)):
            return self.get_response(request)

        try:
            visitor = Visitor.objects.get(
                uuid=visitor_uuid,
                is_active=True,
            )
        except Visitor.DoesNotExist:
            session.clear_visitor_uuid(request)
            return self.get_response(request)
        else:
            request.visitor = visitor
            request.user.is_visitor = True

        return self.get_response(request)


class VisitorDebugMiddleware:
    """Print out visitor info - DEBUG only."""

    def __init__(self, get_response: Callable):
        if not settings.DEBUG:
            raise MiddlewareNotUsed("VisitorDebugMiddleware disabled")
        self.get_response = get_response

    def __call__(self, request: HttpRequest) -> HttpResponse:
        logger.debug("request.user.is_visitor: %s", request.user.is_visitor)
        if request.user.is_visitor:
            logger.debug("request.visitor: %s", request.visitor)
        return self.get_response(request)



from __future__ import annotations

import json

from django.contrib import admin, messages
from django.db.models.query import QuerySet
from django.http.request import HttpRequest
from django.utils.safestring import mark_safe

from .models import Visitor, VisitorLog


def pretty_print(data: dict | None) -> str:
    """Convert dict into formatted HTML."""
    if data is None:
        return ""
    pretty = json.dumps(data, sort_keys=True, indent=4, separators=(",", ": "))
    html = pretty.replace(" ", "&nbsp;").replace("\n", "<br>")
    return mark_safe("<pre><code>%s</code></pre>" % html)


@admin.register(Visitor)
class VisitorsAdmin(admin.ModelAdmin):
    """Admin model for Visitor objects."""

    def deactivate(self, request: HttpRequest, queryset: QuerySet) -> None:
        """Call deactivate on all selected Visitor objects."""
        count = queryset.count()
        for obj in queryset:
            obj.deactivate()
        self.message_user(
            request, f"{count} passes have been disabled.", messages.SUCCESS
        )

    deactivate.short_description = "Deactivate selected Visitor passes"  # type: ignore

    def reactivate(self, request: HttpRequest, queryset: QuerySet) -> None:
        """Reactivate all selected Visitor objects."""
        count = queryset.count()
        for obj in queryset:
            obj.reactivate()
        self.message_user(
            request, f"{count} passes have been activated.", messages.SUCCESS
        )

    reactivate.short_description = "Reactivate selected Visitor passes"  # type: ignore

    actions = (deactivate, reactivate)
    list_filter = ("scope",)
    list_display = (
        "scope",
        "email",
        "created_at",
        "expires_at",
        "is_active",
        "_is_valid",
    )
    readonly_fields = (
        "uuid",
        "created_at",
        "last_updated_at",
        "_context",
        "is_active",
        "expires_at",
    )
    search_fields = (
        "first_name",
        "last_name",
        "email",
        "scope",
        "created_at",
    )

    def _is_valid(self, obj: Visitor) -> bool:
        return obj.is_valid

    _is_valid.boolean = True  # type: ignore

    def _context(self, obj: Visitor) -> str:
        return pretty_print(obj.context)

    _context.short_description = "Context (prettified)"  # type: ignore


@admin.register(VisitorLog)
class VisitorLogAdmin(admin.ModelAdmin):
    list_display = (
        "visitor",
        "session_key",
        "remote_addr",
        "request_uri",
        "status_code",
        "timestamp",
    )
    readonly_fields = [f.name for f in VisitorLog._meta.fields]
    pass



from __future__ import annotations

import uuid

from django.http import HttpRequest, HttpResponse
from django.http.response import HttpResponseRedirect
from django.shortcuts import get_object_or_404, render
from django.urls import reverse
from django.views import View

from visitors.exceptions import InvalidVisitorPass

from .forms import SelfServiceForm
from .models import Visitor
from .signals import self_service_visitor_created


class SelfService(View):
    """
    Enable users to create their own visitor passes.

    If a visitor pass is marked as `auto-enroll` then a user can
    create their own pass. In this model, an user without access
    would be redirected to this view, where they can input their
    own name and email. They receive their own visitor pass.

    The purpose of this is to add some protection to 'obscure'
    urls. e.g. those that rely on an unguessable uuid - instead
    of making those URLs completely public, we can force the user
    to give us their name / email (name can be faked, but they
    will require access to the email).

    """

    def get(self, request: HttpRequest, visitor_uuid: uuid.UUID) -> HttpResponse:
        """Render the initial form."""
        visitor = get_object_or_404(Visitor, uuid=visitor_uuid)
        if visitor.is_active:
            raise InvalidVisitorPass("Visitor pass has already been activated")
        if visitor.has_expired:
            raise InvalidVisitorPass("Visitor pass has expired")
        form = SelfServiceForm(initial={"vuid": visitor.uuid})
        return render(
            request,
            template_name="visitors/self_service_request.html",
            context={"visitor": visitor, "form": form},
        )

    def post(self, request: HttpRequest, visitor_uuid: uuid.UUID) -> HttpResponse:
        """
        Process the form and send the pass email.

        The core function here is to update the visitor object with
        the details from the form. Once that is done the visitor
        pass is active and can be sent to the user. This view fires
        the `self_service_visitor_created` signal - you should use
        this to send out the notification.

        """
        visitor = get_object_or_404(Visitor, uuid=visitor_uuid)
        form = SelfServiceForm(request.POST)
        if form.is_valid():
            visitor.first_name = form.cleaned_data["first_name"]
            visitor.last_name = form.cleaned_data["last_name"]
            visitor.email = form.cleaned_data["email"]
            visitor.reactivate()
            # hook into this to send the email notification to the user.
            self_service_visitor_created.send(sender=self.__class__, visitor=visitor)
            url = reverse(
                "visitors:self-service-success",
                kwargs={"visitor_uuid": visitor_uuid},
            )
            return HttpResponseRedirect(url)
        return render(
            request,
            template_name="visitors/self_service_request.html",
            context={
                "visitor": visitor,
                "form": form,
            },
        )


def self_service_success(request: HttpRequest, visitor_uuid: uuid.UUID) -> HttpResponse:
    """Display the success page."""
    visitor = get_object_or_404(Visitor, uuid=visitor_uuid)
    return render(
        request,
        template_name="visitors/self_service_success.html",
        context={"visitor": visitor},
    )



from __future__ import annotations

import datetime
import uuid
from typing import Any
from urllib.parse import parse_qs, urlencode, urlparse, urlunparse

from django.db import models
from django.db.models.deletion import CASCADE
from django.http.request import HttpRequest
from django.utils.timezone import now as tz_now
from django.utils.translation import gettext_lazy as _lazy

from .exceptions import InvalidVisitorPass
from .settings import VISITOR_QUERYSTRING_KEY, VISITOR_TOKEN_EXPIRY


class VisitorManager(models.Manager):
    def create_temp_visitor(self, scope: str, redirect_to: str) -> Visitor:
        """Create empty Visitor object for self-service."""
        return self.create(
            email=Visitor.DEFAULT_SELF_SERVICE_EMAIL,
            scope=scope,
            is_active=False,
            context={"self-service": True, "redirect_to": redirect_to},
        )


class Visitor(models.Model):
    """A temporary visitor (betwixt anonymous and authenticated)."""

    DEFAULT_TOKEN_EXPIRY = datetime.timedelta(seconds=VISITOR_TOKEN_EXPIRY)
    DEFAULT_SELF_SERVICE_EMAIL = "anon@example.com"

    uuid = models.UUIDField(default=uuid.uuid4, unique=True)
    first_name = models.CharField(max_length=150, blank=True)
    last_name = models.CharField(max_length=150, blank=True)
    email = models.EmailField(db_index=True)
    scope = models.CharField(
        max_length=100, help_text=_lazy("Used to map request to view function")
    )
    created_at = models.DateTimeField(default=tz_now)
    context = models.JSONField(
        null=True,
        blank=True,
        help_text=_lazy("Used to store arbitrary contextual data."),
    )
    last_updated_at = models.DateTimeField(auto_now=True)
    expires_at = models.DateTimeField(
        blank=True,
        null=True,
        help_text=_lazy(
            "After this time the link can no longer be used - "
            "defaults to VISITOR_TOKEN_EXPIRY."
        ),
    )
    is_active = models.BooleanField(
        default=True,
        help_text=_lazy(
            "Set to False to disable the visitor link and prevent further access."
        ),
    )

    objects = VisitorManager()

    class Meta:
        verbose_name = "Visitor pass"
        verbose_name_plural = "Visitor passes"

    def __str__(self) -> str:
        return f"Visitor pass {self.id} (scope='{self.scope}')"

    def __repr__(self) -> str:
        return (
            f"<Visitor id={self.id} uuid='{self.uuid}' "
            f"email='{self.email}' scope='{self.scope}'>"
        )

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        if not self.expires_at:
            self.expires_at = self.created_at + self.DEFAULT_TOKEN_EXPIRY

    @property
    def full_name(self) -> str:
        return f"{self.first_name} {self.last_name}"

    @property
    def session_data(self) -> str:
        return str(self.uuid)

    @property
    def has_expired(self) -> bool:
        """Return True if the expires_at timestamp has been passed (or not yet set)."""
        if not self.expires_at:
            return False
        return self.expires_at < tz_now()

    @property
    def is_valid(self) -> bool:
        """Return True if the token is active and not yet expired."""
        return self.is_active and not self.has_expired

    def is_self_service(self) -> bool:
        """Return True if the token was a self-service token."""
        return self.context.get("self-service", False)

    def validate(self) -> None:
        """Raise InvalidVisitorPass if inactive or expired."""
        if not self.is_active:
            raise InvalidVisitorPass("Visitor pass is inactive")
        if self.has_expired:
            raise InvalidVisitorPass("Visitor pass has expired")

    def serialize(self) -> dict:
        """
        Return JSON-serializable representation.

        Useful for template context and session data.

        """
        return {
            "uuid": str(self.uuid),
            "first_name": self.first_name,
            "last_name": self.last_name,
            "full_name": self.full_name,
            "email": self.email,
            "scope": self.scope,
            "context": self.context,
        }

    def tokenise(self, url: str) -> str:
        """Combine url with querystring token."""
        # from https://stackoverflow.com/a/2506477/45698
        parts = list(urlparse(url))
        query = parse_qs(parts[4])
        query.update({VISITOR_QUERYSTRING_KEY: self.uuid})
        parts[4] = urlencode(query)
        return urlunparse(parts)

    def deactivate(self) -> None:
        """Deactivate the token so it can no longer be used."""
        self.is_active = False
        self.save()

    def reactivate(self) -> None:
        """Reactivate the token so it can be reused."""
        self.is_active = True
        self.expires_at = tz_now() + self.DEFAULT_TOKEN_EXPIRY
        self.save()


class VisitorLogManager(models.Manager):
    def create_log(self, request: HttpRequest, status_code: int) -> VisitorLog:
        """Extract values from HttpRequest and store locally."""
        return self.create(
            visitor=request.visitor,
            session_key=request.session.session_key or "",
            http_method=request.method,
            request_uri=request.path,
            query_string=request.META.get("QUERY_STRING", ""),
            http_user_agent=request.META.get("HTTP_USER_AGENT", ""),
            # we care about the domain more than the URL itself, so truncating
            # doesn't lose much useful information
            http_referer=request.META.get("HTTP_REFERER", ""),
            # X-Forwarded-For is used by convention when passing through
            # load balancers etc., as the REMOTE_ADDR is rewritten in transit
            remote_addr=(
                request.META.get("HTTP_X_FORWARDED_FOR")
                if "HTTP_X_FORWARDED_FOR" in request.META
                else request.META.get("REMOTE_ADDR")
            ),
            status_code=status_code,
        )


class VisitorLog(models.Model):
    """Log visitors."""

    visitor = models.ForeignKey(Visitor, related_name="visits", on_delete=CASCADE)
    session_key = models.CharField(blank=True, max_length=40)
    http_method = models.CharField(max_length=10)
    request_uri = models.URLField()
    remote_addr = models.CharField(max_length=100)
    query_string = models.TextField(blank=True)
    http_user_agent = models.TextField()
    http_referer = models.TextField()
    status_code = models.PositiveIntegerField("HTTP Response", default=0)
    timestamp = models.DateTimeField(default=tz_now)

    objects = VisitorLogManager()



from django import forms
from django.core.exceptions import ValidationError
from django.utils.translation import gettext as _


class SelfServiceForm(forms.Form):
    """Form for capturing user info."""

    first_name = forms.CharField(max_length=150)
    last_name = forms.CharField(max_length=150)
    email = forms.EmailField(required=True)
    vuid = forms.CharField(widget=forms.HiddenInput())

    def clean_email(self) -> str:
        email = self.cleaned_data["email"]
        if email.endswith("example.com"):
            raise ValidationError(
                _("Invalid email - you cannot use the example.com domain")
            )
        return email



from django.http.request import HttpRequest

from visitors.settings import VISITOR_SESSION_EXPIRY, VISITOR_SESSION_KEY


def stash_visitor_uuid(request: HttpRequest) -> None:
    """Store request visitor data in session."""
    request.session[VISITOR_SESSION_KEY] = request.visitor.session_data
    if request.user.is_anonymous:
        request.session.set_expiry(VISITOR_SESSION_EXPIRY)


def get_visitor_uuid(request: HttpRequest) -> str:
    """Return visitor data from session."""
    return request.session.get(VISITOR_SESSION_KEY, "")


def clear_visitor_uuid(request: HttpRequest) -> None:
    """Remove visitor data from session."""
    request.session.pop(VISITOR_SESSION_KEY, "")



# Generated by Django 3.1.6 on 2021-02-13 15:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("visitors", "0004_visitor_expires_at"),
    ]

    operations = [
        migrations.AddField(
            model_name="visitorlog",
            name="status_code",
            field=models.PositiveIntegerField(default=0, verbose_name="HTTP Response"),
        ),
    ]



# Generated by Django 3.1.4 on 2020-12-23 17:10

import uuid

import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Visitor",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("uuid", models.UUIDField(default=uuid.uuid4)),
                ("first_name", models.CharField(blank=True, max_length=150)),
                ("last_name", models.CharField(blank=True, max_length=150)),
                ("email", models.EmailField(db_index=True, max_length=254)),
                (
                    "scope",
                    models.CharField(
                        help_text="Used to map request to view function", max_length=100
                    ),
                ),
                ("created_at", models.DateTimeField(default=django.utils.timezone.now)),
                (
                    "context",
                    models.JSONField(
                        blank=True,
                        help_text="Used to store arbitrary contextual data.",
                        null=True,
                    ),
                ),
                ("last_updated_at", models.DateTimeField(auto_now=True)),
            ],
            options={
                "verbose_name": "Visitor pass",
                "verbose_name_plural": "Visitor passes",
            },
        ),
    ]



# Generated by Django 3.1.4 on 2021-01-03 12:29

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("visitors", "0002_visitorlog"),
    ]

    operations = [
        migrations.AddField(
            model_name="visitor",
            name="is_active",
            field=models.BooleanField(
                default=True,
                help_text="Set to False to disable the visitor link and prevent further access.",
            ),
        ),
    ]



# Generated by Django 3.1.4 on 2021-01-03 12:37

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("visitors", "0003_visitor_is_active"),
    ]

    operations = [
        migrations.AddField(
            model_name="visitor",
            name="expires_at",
            field=models.DateTimeField(
                blank=True,
                help_text="After this time the link can no longer be used - defaults to VISITOR_TOKEN_EXPIRY.",
                null=True,
            ),
        ),
    ]



from __future__ import annotations

from django.http import HttpRequest


def visitor(request: HttpRequest) -> dict[str, object]:
    """Add visitor context to template context (if found on the request)."""
    if not request.visitor:
        return {"visitor": None}
    return {"visitor": request.visitor.serialize()}



# Generated by Django 3.1.4 on 2020-12-23 18:42

import django.db.models.deletion
import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("visitors", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="VisitorLog",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("session_key", models.CharField(blank=True, max_length=40)),
                ("http_method", models.CharField(max_length=10)),
                ("request_uri", models.URLField()),
                ("remote_addr", models.CharField(max_length=100)),
                ("query_string", models.TextField(blank=True)),
                ("http_user_agent", models.TextField()),
                ("http_referer", models.TextField()),
                ("timestamp", models.DateTimeField(default=django.utils.timezone.now)),
                (
                    "visitor",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="visits",
                        to="visitors.visitor",
                    ),
                ),
            ],
        ),
    ]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['visitors', 'visitors.migrations']

package_data = \
{'': ['*'], 'visitors': ['templates/visitors/*']}

install_requires = \
['django>=3.0,<4.0']

setup_kwargs = {
    'name': 'django-visitor-pass',
    'version': '0.5',
    'description': 'Django app for managing temporary session-based users.',
    'long_description': '# Django Visitor Pass\n\nDjango app for managing temporary session-based users.\n\n### Background\n\nThis package has been extracted out of `django-request-token` as a specific use\ncase - that of temporary site "visitors". It enables a type of ephemeral user\nwho is neither anonymous nor authenticated, but somewhere in between - known for\nthe duration of their session.\n\n### Motivation\n\nWe\'ve been using `django-request-token` for a while, and have issued over\n100,000 tokens. A recent analysis showed two main use cases - single-use "magic\nlinks" for logging people in, and a more involved case where we invite\nunregistered users on to the platform to perform some action - providing a\nreference perhaps, or collaborating on something with (registered) users. The\nformer we have extracted out into `django-magic-links` - and this package\naddresses the latter.\n\n### What is a "visitor"?\n\nIn the standard Django model you have the concept of an `AnonymousUser`, and an\nauthenticated `User` - someone who has logged in. We have a third, intermediate,\ntype of user - which we have historically referred to as a "Temp User", which is\nsomeone we know _of_, but who has not yet registered.\n\nThe canonical example of this is leaving a reference: user A on the site invites\nuser B to leave a reference for them. They (A) give us B\'s name and email, we\ninvite them to click on a link and fill out a form. That\'s it. We store their\nname and email so that we can contact them, but it\'s ephemeral - we don\'t need\nit, and we don\'t use it. Storing this data in the User table made sense (as it\nhas name and email fields), but it led to a lot of `user_type=TEMP` munging to\ndetermine who is a \'real\' user on the site.\n\nWhat we really want is to \'stash\' this information somewhere outside of the auth\nsystem, and to enable these temp users to have restricted access to specific\nareas of the application, for a limited period, after which we can forget about\nthem and clear out the data.\n\nWe call these users "visitors".\n\n### Use Case - request a reference\n\nFred is a registered user on the site, and would like a reference from Ginger,\nhis dance partner.\n\n1. Fred fills out the reference request form:\n\n```\n   Name: Ginger\n   Email: ginger@[...].com\n   Message: ...\n   Scope: REFERENCE_REQUEST [hidden field]\n```\n\n2. We save this information, and generate a unique link which we send to Ginger,\n   along with the message.\n\n3. Ginger clicks on the link, at which point we recognise that this is someone\n   we know about - a "visitor" - but who is in all other respects an\n   `AnonymousUser`.\n\n4. We stash the visitor info in the standard session object - so that even\n   though Ginger is not authenticated, we know who she is, and more importantly\n   we know why she\'s here (REFERENCE_REQUEST).\n\n5. Ginger submits the reference - which may be a multi-step process, involving\n   GETs and POSTs, all of which are guarded by a decorator that restricts access\n   to visitors with the appropriate Scope (just like `django-request-token`).\n\n6. At the final step we can (optionally) choose to clear the session info\n   immediately, effectively removing all further access.\n\n### Implementation\n\nThis code has been extracted out of `django-request-token` and simplified. It\nstores the visitor data in the `Visitor` model, and on each successful first\nrequest (where the token is \'processed\' and the session filled) we record a\n`VisitorLog` record. This includes HTTP request info (session_key, referer,\nclient IP, user-agent). This information is for analytics only - for instance\ndetermining whether links are being shared.\n\nThe app works by adding some attributes to the `request` and `request.user`\nobjects. The user has a boolean `user.is_visitor` property, and the request has\na `request.visitor` property which is the relevant `Visitor` object.\n\nThis is done via two bits of middleware, `VisitorRequestMiddleware` and\n`VisitSessionMiddleware`. \n\n#### `VisitorRequestMiddleware`\n\nThis middleware looks for a visitor token (uuid) on the incoming request\nquerystring. If it finds a token, it will look up the `Visitor` object, add it\nto the request, and then set the `request.user.is_visitor` attribute. It sets\nthe properties from the request, and has no interaction with the session. This\nhappens in the second piece of middleware.\n\n#### `VisitorSessionMiddleware`\n\nThis middleware must come after the `VisitorRequestMiddleware` (it will blow up\nif it can\'t access `request.visitor`). It has two responsibilities:\n\n1. If the request object has a visitor object on it, then it _must_ have been\n   set by the request middleware on the current request - so it\'s a new visitor,\n   and we immediately stash it in the `request.session`. \n\n1. If `request.visitor` is None, then we don\'t have a _new_ visitor, but there\n   may be one already stashed in the `request.session`, in which case we want to\n   add it on the to the request.\n\nNote: splitting this in two seems over-complicated, but because we are moving\nvalues from request-into-session-into-request it\'s a lot simpler to run two\ncompletely separate passes.\n\n### Configuration\n\n#### Django Settings\n\n1. Add `visitors` to `INSTALLED_APPS`\n1. Add `visitors.middleware.VisitorRequestMiddleware` to `MIDDLEWARE`\n1. Add `visitors.middleware.VisitorSessionMiddleware` to `MIDDLEWARE`\n\n#### Environment Settings\n\n* `VISITOR_SESSION_KEY`: session key used to stash visitor info (defaut:\n  `visitor:session`)\n\n* `VISITOR_QUERYSTRING_KEY`: querystring param used on tokenised links (default:\n  `vuid`)\n\n### Usage\n\nOnce you have the package configured, you can use the `user_is_visitor`\ndecorator to protect views that you want to restricted to visitors only. The\ndecorator requires a `scope` kwarg, which must match the `Visitor.scope` value\nset when the `Visitor` object is created.\n\n```python\n@user_is_visitor(scope="foo")\ndef protected_view(request):\n   pass\n```\n\nBy default the decorator will allow visitors with the correct scope only. If you\nwant more control over the access, you can pass a callable as the `bypass_func`\nparam:\n\n```python\n# allow authenticated users as well as visitors\n@user_is_visitor(scope="foo", bypass_func=lambda r: r.user.is_staff)\ndef allow_visitors_and_staff(request):\n   pass\n```\n\nIf you don\'t care about the scope (you should), you can use `"*"` to allow all\nvisitors access:\n\n```python\n@user_is_visitor(scope="*")  # see also SCOPE_ANY\ndef allow_all_scopes(request):\n   pass\n```\n\nAlternatively, for more complex use cases, you can ignore the decorator and just\ninspect the request itself:\n\n```python\ndef complicated_rules(request):\n   if request.user.is_visitor:\n      pass\n   elif is_national_holiday():\n      pass\n   elif is_sunny_day():\n      pass\n   else:\n      raise PermissionDenied\n```\n',
    'author': 'YunoJuno',
    'author_email': 'code@yunojuno.com',
    'maintainer': 'YunoJuno',
    'maintainer_email': 'code@yunojuno.com',
    'url': 'https://github.com/yunojuno/django-visitor-pass',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from django.core.exceptions import PermissionDenied


class InvalidVisitorPass(Exception):
    pass


class VisitorAccessDenied(PermissionDenied):
    """Error raised by decorator when visitor is invalid."""

    def __init__(self, msg: str, scope: str) -> None:
        super().__init__(msg)
        self.scope = scope



from __future__ import annotations

import functools
import logging
from typing import Any, Callable

from django.conf import settings
from django.http import HttpRequest, HttpResponse
from django.http.response import HttpResponseRedirect
from django.urls import reverse
from django.utils.translation import gettext as _

from .exceptions import VisitorAccessDenied
from .models import Visitor, VisitorLog

logger = logging.getLogger(__name__)

# universal scope - essentially unscoped access
SCOPE_ANY = "*"

# for typing
BypassFunc = Callable[[HttpRequest], bool]


def is_visitor(user: settings.AUTH_USER_MODEL) -> bool:
    """Shortcut function for use with user_passes_test decorator."""
    return user.is_visitor


def is_staff(user: settings.AUTH_USER_MODEL) -> bool:
    """Shortcut function for use with user_passes_test decorator."""
    return user.is_staff


def is_superuser(user: settings.AUTH_USER_MODEL) -> bool:
    """Shortcut function for use with user_passes_test decorator."""
    return user.is_superuser


def is_authenticated(user: settings.AUTH_USER_MODEL) -> bool:
    """Shortcut function for use with user_passes_test decorator."""
    return user.is_authenticated


def _get_request_arg(*args: Any) -> HttpRequest | None:
    """Extract the arg that is an HttpRequest object."""
    for arg in args:
        if isinstance(arg, HttpRequest):
            return arg
    return None


def user_is_visitor(  # noqa: C901
    view_func: Callable | None = None,
    scope: str = "",
    bypass_func: BypassFunc | None = None,
    log_visit: bool = True,
    self_service: bool = False,
) -> Callable:
    """
    Decorate view functions that supports Visitor access.

    The 'scope' param is mapped to the request.visitor.scope attribute - if
    the scope is SCOPE_ANY then this is ignored.

    The 'bypass_func' is a callable that can be used to provide exceptions
    to the scope - e.g. allowing authenticate users, or staff, to bypass the
    visitor restriction. Defaults to None (only visitors with appropriate
    scope allowed).

    The 'log_visit' arg can be used to override the default logging - if this
    is too noisy, for instance.

    If 'self_service' is True, then instead of a straight PermissionDenied error
    we raise VisitorAccessDenied, passing along the scope. This is then picked
    up in the middleware, and the user redirected to a page where they can
    enter their details and effectively invite themselves. Caveat emptor.

    """
    if not scope:
        raise ValueError("Decorator scope cannot be empty.")

    if view_func is None:
        return functools.partial(
            user_is_visitor,
            scope=scope,
            bypass_func=bypass_func,
            log_visit=log_visit,
            self_service=self_service,
        )

    @functools.wraps(view_func)
    def inner(*args: Any, **kwargs: Any) -> HttpResponse:
        # should never happen, but keeps mypy happy as it _could_
        if not view_func:
            raise ValueError("Callable (view_func) missing.")

        # HACK: if this is decorating a method, then the first arg will be
        # the object (self), and not the request. In order to make this work
        # with functions and methods we need to determine where the request
        # arg is.
        request = _get_request_arg(*args)
        if not request:
            raise ValueError("Request argument missing.")

        # Allow custom rules to bypass the visitor checks
        if bypass_func and bypass_func(request):
            return view_func(*args, **kwargs)

        if not is_valid_request(request, scope):
            if self_service:
                return redirect_to_self_service(request, scope)
            raise VisitorAccessDenied(_("Visitor access denied"), scope)

        response = view_func(*args, **kwargs)
        if log_visit:
            VisitorLog.objects.create_log(request, response.status_code)
        return response

    return inner


def is_valid_request(request: HttpRequest, scope: str) -> bool:
    """Return True if the request matches the scope."""
    if not request.user.is_visitor:
        return False
    if scope == SCOPE_ANY:
        return True
    return request.visitor.scope == scope


def redirect_to_self_service(request: HttpRequest, scope: str) -> HttpResponseRedirect:
    """Create inactive Visitor token and redirect to enable self-service."""
    # create an inactive token for the time being. This will be used by
    # the auto-enroll view. The user fills in their name and email, which
    # overwrites the blank values here, and sets the token to be active.
    visitor = Visitor.objects.create_temp_visitor(
        scope=scope, redirect_to=request.get_full_path()
    )
    return HttpResponseRedirect(
        reverse(
            "visitors:self-service",
            kwargs={"visitor_uuid": visitor.uuid},
        )
    )



from __future__ import annotations

from django.conf import settings


def _setting(key: str, default: str) -> str:
    return getattr(settings, key, default)


# session key used to store visitor.uuid
VISITOR_SESSION_KEY: str = _setting("VISITOR_SESSION_KEY", "visitor:session")

# key used to store visitor uuid on querystring
VISITOR_QUERYSTRING_KEY: str = _setting("VISITOR_QUERYSTRING_KEY", "vuid")

# Value used to `request.session.set_expiry` for anonymous users. If the
# request.user is already authenticated we let them carry on with the site
# default - however if the visitor is anonymous we can override the session
# expiry. By default this will be set to 0, which means that the session expires
# when the browser is closed. We do not support datetime or timedelta values
# here - it must be an integer (seconds), or None. see
# https://docs.djangoproject.com/en/3.1/topics/http/sessions/#django.contrib.sessions.backends.base.SessionBase.set_expiry  # noqa
# * If value is an integer, the session will expire after that many seconds of
#   inactivity.
# * If value is None, the session reverts to using the global session expiry
#   policy.
VISITOR_SESSION_EXPIRY: int | None = _setting("VISITOR_SESSION_EXPIRY", 0)

# Value used to set the expiry of the visitor token - the point at which it can
# no longer be used. NB this is separate from the session expiry. Once the token
# is stashed in the session the visitor will remain a visitor until the session
# expires. This value is used by the VisitorRequestMiddleware.
VISITOR_TOKEN_EXPIRY: int = _setting("VISITOR_TOKEN_EXPIRY", 300)
